from torch import nn

from .bot_att import BottleneckAttention
from ..axial_attention_deeplab.axial_attention_residual_block import _conv2d1x1


class BottleneckBlock(nn.Module):
    def __init__(self, *, in_channels,
                 fmap_size,
                 out_channels=2048,
                 proj_factor=4,
                 heads=4,
                 dim_head=None,
                 pooling=False,
                 content_positional_embedding=True):
        """
        paper: https://arxiv.org/abs/2101.11605
        Check figure 3, and Table 1

        Args:
            in_channels: number of feat_maps
            fmap_size: spatial dims
            out_channels: the desired output channels of the layers
            proj_factor: used to calc. the bottleneck dim of the MHSA. (out_channels // proj_factor)
            heads: number of representation to learn from the input
            dim_head: defaults to (bottleneck_dimension/heads)
            pooling: (bool) whether to apply avg pool after bot-MHSA
            content_positional_embedding: (bool) whether to apply 2D rel pos enc
        """
        super().__init__()
        bottleneck_dimension = out_channels // proj_factor  # contraction_channels 512
        mhsa_out_channels = bottleneck_dimension if dim_head is None else dim_head * heads

        contraction = _conv2d1x1(in_channels, bottleneck_dimension)

        bot_mhsa = BottleneckAttention(
            dim=bottleneck_dimension,
            fmap_size=fmap_size,
            heads=heads,
            dim_head=dim_head,
            content_positional_embedding=content_positional_embedding)

        pool_layer = nn.AvgPool2d(kernel_size=(2, 2), stride=(2, 2)) if pooling else nn.Identity()

        expansion = _conv2d1x1(mhsa_out_channels, out_channels)

        self.block = nn.Sequential(contraction,
                                   nn.ReLU(),
                                   bot_mhsa,
                                   pool_layer,
                                   nn.BatchNorm2d(mhsa_out_channels),
                                   nn.ReLU(),
                                   expansion)  # no relu after expansion

        # TODO find init_zero=True tf param for batch norm

        # skip connection
        if pooling or in_channels != out_channels:
            # if we have pooling we need a 2-strided 1x1 conv
            stride = 2 if pooling else 1
            # we also need to match the desired out_channels from the Bot-MHSA
            self.shortcut = nn.Sequential(
                _conv2d1x1(in_channels, out_channels, stride=stride),
                nn.ReLU())
        else:
            self.shortcut = nn.Identity()

    def forward(self, x):
        print('forward..', x.shape)
        return self.block(x) + self.shortcut(x)


class BottleneckModule(nn.Module):
    def __init__(self, *, in_channels,
                 fmap_size,
                 out_channels=2048,
                 proj_factor=4,
                 heads=4,
                 dim_head=None,
                 pooling=True,
                 content_positional_embedding=True,
                 num_layers=3,  # default
                 ):
        """
        Applies 3 bottleneck layers as in the original paper
        Args:
            in_channels: number of feat_maps
            fmap_size: spatial dims
            out_channels: the desired output channels of the layers
            proj_factor: used to calc. the bottleneck dim of the MHSA. (out_channels // proj_factor)
            heads: number of representation to learn from the input
            dim_head: defaults to (bottleneck_dimension/heads)
            pooling: (bool) whether to apply avg pool after bot-MHSA
            content_positional_embedding: (bool) whether to apply 2D rel pos enc
            num_layers: 3 used as in paper
        """

        super().__init__()
        block_list = []
        for i in range(num_layers):
            if i == 0:
                feat_map = fmap_size
                if pooling:
                    pool = True
            else:
                pool = False
                if pooling:
                    in_channels = out_channels
                    feat_map = (fmap_size[0] // 2, fmap_size[1] // 2)

            block_list.append(BottleneckBlock(in_channels=in_channels,
                                              fmap_size=feat_map,
                                              out_channels=out_channels,
                                              proj_factor=proj_factor,
                                              heads=heads,
                                              dim_head=dim_head,
                                              pooling=pool,
                                              content_positional_embedding=content_positional_embedding))
        self.model = nn.Sequential(*block_list)

    def forward(self, x):
        return self.model(x)



import torch
from einops import rearrange
from torch import nn

from self_attention_cv.linformer.linformer import project_vk_linformer
from self_attention_cv.transformer_vanilla.mhsa import compute_mhsa
from ..common import expand_to_batch


def split_cls(x):
    """
    split the first token of the sequence in dim 2
    """
    # Note by indexing the first element as 0:1 the dim is kept in the tensor's shape
    return x[:, :, 0:1, ...], x[:, :, 1:, ...]


def time_att_rearrange(x, frames):
    return rearrange(x, 'b h (f p) d -> (b p) h f d', f=frames)


def space_att_rearrange(x, patches):
    return rearrange(x, 'b h (f p) d -> (b f) h p d', p=patches)


def merge_timespace(x, batch, space=False):
    out_indices = 'b h (k t) d' if space else 'b h (t k) d'
    return rearrange(x, f'(b k) h t d -> {out_indices}', b=batch)


class SpacetimeMHSA(nn.Module):
    def __init__(self, dim, tokens_to_attend, space_att, heads=8,
                 dim_head=None, classification=True,
                 linear_spatial_attention=False, k=None):
        """
        Attention through time and space to process videos
        choose mode (whether to operate in space and time with space_att (bool) )
        CLS token is used for video classification, which will attend all tokens in both
        space and time before attention only in time or space.

        Code is based on lucidrains repo: https://github.com/lucidrains/TimeSformer-pytorch
        Args:
            dim: token's dimension, i.e. word embedding vector size
            tokens_to_attend: space (patches) or time (frames) tokens that we will attend to
            heads: the number of distinct representations to learn
            dim_head: the dim of the head. In general dim_head=dim//heads.
            space_att: whether to use space or time attention in this block
            classification: when True a classification token is expected in the forward call
        """
        super().__init__()
        self.dim_head = (int(dim / heads)) if dim_head is None else dim_head
        _dim = self.dim_head * heads
        self.heads = heads
        self.to_qvk = nn.Linear(dim, _dim * 3, bias=False)
        self.W_0 = nn.Linear(_dim, dim, bias=False)
        self.scale_factor = self.dim_head ** -0.5
        self.space_att = space_att
        self.reshape_timespace = space_att_rearrange if self.space_att else time_att_rearrange
        self.tokens_to_attend = tokens_to_attend
        self.classification = classification
        self.linear_spatial_attention = linear_spatial_attention and self.space_att
        self.k = k if k is not None else 256

        if self.linear_spatial_attention:
            proj_shape = tuple((self.tokens_to_attend + 1, k))
            self.E = torch.nn.Parameter(torch.randn(proj_shape))

    def forward(self, x):
        """
        Expects input x with merged tokens in both space and time
        Args:
            x: [batch, tokens_timespace+ cls_token, dim*3*heads ]
        """
        assert x.dim() == 3
        batch, token_dim = x.shape[0], 2
        qkv = self.to_qvk(x)

        # decomposition to q,v,k and cast to tuple
        q, k, v = tuple(rearrange(qkv, 'b t (d k h ) -> k b h t d ', k=3, h=self.heads))

        if self.classification:
            # only the cls token will to ALL tokens in both space+time: tokens_timespace
            (cls_q, q_3D) = split_cls(q)
            out_cls = compute_mhsa(cls_q, k, v, scale_factor=self.scale_factor)

            # reshape for space or time attention here only
            (cls_k, k_3D), (cls_v, v_3D) = map(split_cls, (k, v))

            # this is where we decompose/separate the tokens for attention in time or space only
            q_sep, k_sep, v_sep = map(self.reshape_timespace, [q_3D, k_3D, v_3D], [self.tokens_to_attend] * 3)

            # we have to expand/repeat the cls_k, and cls_v to k,v
            cls_k, cls_v = map(expand_to_batch, (cls_k, cls_v), (k_sep.shape[0], v_sep.shape[0]))

            k = torch.cat((cls_k, k_sep), dim=token_dim)
            v = torch.cat((cls_v, v_sep), dim=token_dim)

            if self.linear_spatial_attention:
                v, k = project_vk_linformer(v, k, self.E)

            # finally the conventional attention only through space/time
            out_mhsa = compute_mhsa(q_sep, k, v, scale_factor=self.scale_factor)

            # merge tokens from space and time
            out_mhsa = merge_timespace(out_mhsa, batch, self.space_att)
            # and spacetime cls token
            out = torch.cat((out_cls, out_mhsa), dim=token_dim)
        else:
            out = compute_mhsa(q, k, v)

        # re-compose: merge heads with dim_head
        out = rearrange(out, "b h t d -> b t (h d)")
        # Apply final linear transformation layer
        return self.W_0(out)



import torch.nn as nn


# taken and adapted form torchvision repo
# https://github.com/pytorch/vision/blob/af97ec2f4c9daac091b9a87355c4f22d37488004/torchvision/models/resnet.py#L86

def conv3x3(in_planes: int, out_planes: int, stride: int = 1, groups: int = 1, dilation: int = 1) -> nn.Conv2d:
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes: int, out_planes: int, stride: int = 1) -> nn.Conv2d:
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class Bottleneck(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, groups=1,
                 base_width=64, dilation=1, norm_layer=None):
        super(Bottleneck, self).__init__()

        if norm_layer is None:
            norm_layer = nn.BatchNorm2d

        if stride != 1 or inplanes != planes * self.expansion:
            self.downsample = nn.Sequential(
                conv1x1(inplanes, planes * self.expansion, stride),
                norm_layer(planes * self.expansion),
            )
        else:
            self.downsample = nn.Identity()

        width = int(planes * (base_width / 64.)) * groups

        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)
        identity = self.downsample(x)
        out += identity
        out = self.relu(out)

        return out



import torch
import torch.nn as nn
from ..common import expand_to_batch


# adapted from https://pytorch.org/tutorials/beginner/transformer_tutorial.html
class PositionalEncodingSin(nn.Module):

    def __init__(self, dim, dropout=0.1, max_tokens=5000):
        super(PositionalEncodingSin, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(1, max_tokens, dim)
        position = torch.arange(0, max_tokens, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, dim, 2).float() * (-torch.log(torch.Tensor([10000.0])) / dim))
        pe[..., 0::2] = torch.sin(position * div_term)
        pe[..., 1::2] = torch.cos(position * div_term)
        #pe = pe.unsqueeze(0).transpose(0, 1)
        self.pe = pe

    def forward(self, x):
        batch, seq_tokens, _ = x.size()
        x = x + expand_to_batch( self.pe[:, :seq_tokens, :], desired_size=batch)
        return self.dropout(x)


import torch
from einops import rearrange
from torch import nn

from self_attention_cv.pos_embeddings.relative_pos_enc_qkv import Relative2DPosEncQKV


def _conv1d1x1(in_channels, out_channels):
    """1D convolution with kernel size of 1 followed by batch norm"""
    return nn.Sequential(nn.Conv1d(in_channels, out_channels, kernel_size=1, stride=1, bias=False),
                         nn.BatchNorm1d(out_channels))


class AxialAttention(nn.Module):
    def __init__(self, dim, in_channels=128, heads=8, dim_head_kq=8):
        """
        Fig.1 page 6 in Axial DeepLab paper

        Args:
            in_channels: the channels of the feature map to be convolved by 1x1 1D conv
            heads: number of heads
            dim_head_kq: inner dim
        """
        super().__init__()
        self.dim_head = in_channels // heads
        self.dim = dim

        self.heads = heads

        self.dim_head_v = self.dim_head  # d_out
        self.dim_head_kq = dim_head_kq
        self.qkv_channels = self.dim_head_v + self.dim_head_kq * 2
        self.to_qvk = _conv1d1x1(in_channels, self.heads * self.qkv_channels)

        # Position embedding 2D
        self.RelativePosEncQKV = Relative2DPosEncQKV(dim, self.dim_head_v, self.dim_head_kq)

        # Batch normalization - not common, but we dont need to scale down the dot products this way
        self.attention_norm = nn.BatchNorm2d(heads * 3)
        self.out_norm = nn.BatchNorm1d(in_channels * 2)

    def forward(self, x_in):
        assert x_in.dim() == 3, 'Ensure your input is 4D: [b * width, chan, height] or [b * height, chan, width]'

        # Calculate position embedding -> [ batch*width , qkv_channels,  dim ]
        qkv = self.to_qvk(x_in)

        qkv = rearrange(qkv, 'b (q h) d -> b h q d ', d=self.dim, q=self.qkv_channels, h=self.heads)

        # dim_head_kq != dim_head_v so I cannot decompose with einops here I think
        q, k, v = torch.split(qkv, [self.dim_head_kq, self.dim_head_kq, self.dim_head_v], dim=2)

        r_q, r_k, r_v = self.RelativePosEncQKV()

        # Computations are carried as Fig.1 page 6 in Axial DeepLab paper
        qr = torch.einsum('b h i d, i d j -> b h d j ', q, r_q)
        kr = torch.einsum('b h i d, i d j -> b h d j ', k, r_k)

        dots = torch.einsum('b h i d, b h i j -> b h d j', q, k)

        # We normalize the 3 tensors qr, kr, dots together before element-wise addition
        # To do so we concatenate the tensor heads just to normalize them
        # conceptually similar to scaled dot product in MHSA
        # Here n = len(list)
        norm_dots = self.attention_norm(rearrange(list([qr, kr, dots]), 'n b h d j -> b (h n) d j'))

        # Now we can decompose them
        norm_dots = rearrange(norm_dots, 'b (h n) d j -> n b h d j', n=3)

        # And use einsum in the n=3 axis for element-wise sum
        norm_dots = torch.einsum('n b h d j -> b h d j', norm_dots)

        # Last dimension is used softmax and matrix multplication
        attn = torch.softmax(norm_dots, dim=-1)
        # Matrix multiplication will be performed in the dimension of the softmax! Attention :)
        out = torch.einsum('b h d j,  b h i j -> b h i d', attn, v)

        # Last embedding of v
        kv = torch.einsum('b h d j, i d j -> b h i d ', attn, r_v)

        # To perform batch norm as described in paper,
        # we will merge the dimensions that are != self.dim
        # n = 2 = len(list)
        out = self.out_norm(rearrange(list([kv, out]), 'n b h i d ->  b (n h i ) d'))
        # decompose back output and merge heads
        out = rearrange(out, 'b (n h i ) d ->  n b (h i) d ', n=2, h=self.heads)
        # element wise sum in n=2 axis
        return torch.einsum('n b j i -> b j i', out)



from .pos_embeddings import *
from .MSA_transformer import MSATransformerBlock, MSATransformerEncoder, TiedRowAxialAttention
from .axial_attention_deeplab import *
from .linformer import LinformerAttention, LinformerBlock, LinformerEncoder
from .timesformer import Timesformer, SpacetimeMHSA
from .transformer_vanilla import *
from .vit import ViT, ResNet50ViT
from .transunet import TransUnet
from .UnetTr import UNETR



from torch import nn

from .mhsa import MultiHeadSelfAttention
from ..common import get_module_device

class TransformerBlock(nn.Module):
    """
    Vanilla transformer block from the original paper "Attention is all you need"
    Detailed analysis: https://theaisummer.com/transformer/
    """

    def __init__(self, dim, heads=8, dim_head=None,
                 dim_linear_block=1024, dropout=0.1, activation=nn.GELU,
                 mhsa=None, prenorm=False):
        """
        Args:
            dim: token's vector length
            heads: number of heads
            dim_head: if none dim/heads is used
            dim_linear_block: the inner projection dim
            dropout: probability of droppping values
            mhsa: if provided you can change the vanilla self-attention block
            prenorm: if the layer norm will be applied before the mhsa or after
        """
        super().__init__()
        self.mhsa = mhsa if mhsa is not None else MultiHeadSelfAttention(dim=dim, heads=heads, dim_head=dim_head)
        self.prenorm = prenorm
        self.drop = nn.Dropout(dropout)
        self.norm_1 = nn.LayerNorm(dim)
        self.norm_2 = nn.LayerNorm(dim)

        self.linear = nn.Sequential(
            nn.Linear(dim, dim_linear_block),
            activation(),  # nn.ReLU or nn.GELU
            nn.Dropout(dropout),
            nn.Linear(dim_linear_block, dim),
            nn.Dropout(dropout)
        )

    def forward(self, x, mask=None):
        if self.prenorm:
            y = self.drop(self.mhsa(self.norm_1(x), mask)) + x
            out = self.linear(self.norm_2(y)) + y
        else:
            y = self.norm_1(self.drop(self.mhsa(x, mask)) + x)
            out = self.norm_2(self.linear(y) + y)
        return out


class TransformerEncoder(nn.Module):
    def __init__(self, dim, blocks=6, heads=8, dim_head=None, dim_linear_block=1024, dropout=0, prenorm=False):
        super().__init__()
        self.block_list = [TransformerBlock(dim, heads, dim_head,
                                            dim_linear_block, dropout, prenorm=prenorm) for _ in range(blocks)]
        self.layers = nn.ModuleList(self.block_list)

    def forward(self, x, mask=None):
        for layer in self.layers:
            x = layer(x, mask)
        return x



from ..transformer_vanilla import MultiHeadSelfAttention, TransformerEncoder

import torch
import torch.nn as nn
from einops import rearrange

from self_attention_cv import TransformerEncoder


class Transformer3dSeg(nn.Module):
    def __init__(self, *,
                 subvol_dim=24,  # W in the paper
                 patch_dim=8,
                 num_classes = 2,
                 in_channels=3,
                 dim=1024,
                 blocks=7,
                 heads=4,
                 dim_linear_block=1024,
                 dropout=0, transformer=None):
        """
        Replication/re-implementation based on the paper:
        Convolution-Free Medical Image Segmentation using Transformers
        Paper: https://arxiv.org/abs/2102.13645
        The model accepts a sub-volume of subvol_dim X subvol_dim X subvol_dim, and return a cube segmentation
        of n = subvol_dim//patch_dim

        Args:
            subvol_dim: cube block size. you have to sample the original 3D volume
            patch_dim: the desired cuded 3D patch dimension
            num_classes: segmentation classes
            in_channels: channels/modalities
            dim: the projected dim of the patches that will be the input dim of the transformer
            blocks: number of repreated transformer blocks
            heads: number of heads
            dim_linear_block: MLP dim inside Transformer block
            dropout: dropout for the transformer block
            transformer: if provided another transformer is used. Vanilla Transformer by default as in ViT
        """
        super().__init__()
        self.p = patch_dim
        self.num_classes = num_classes
        # tokens = number of 3D patches
        self.n = subvol_dim//patch_dim
        self.tokens = self.n ** 3

        self.mid_token = (self.tokens//2)

        self.token_dim = in_channels * (self.p ** 3)
        self.dim = dim

        # Projection and pos embeddings
        self.project_patches = nn.Linear(self.token_dim, dim)

        self.emb_dropout = nn.Dropout(dropout)

        self.pos_emb1D = nn.Parameter(torch.randn(self.tokens, dim))

        if transformer is None:
            self.transformer = TransformerEncoder(dim, blocks=blocks, heads=heads,
                                                  dim_linear_block=dim_linear_block,
                                                  dropout=dropout)
        else:
            self.transformer = transformer

        self.mlp_seg_head = nn.Linear(dim, self.tokens * self.num_classes)

    def forward(self, img, mask=None):
        """
        Args:
            img: square subvolume sub-sampled from the 3D image  b,c,W,W,W
            mask: mask must be supposrted also
        Returns: center patch segmentation
        """
        # Create patches
        # from [batch, channels, h, w ,z ] to [batch, tokens , N], N=p*p*c , tokens = h/p *w/p
        img_patches = rearrange(img,
                                'b c (patch_x x) (patch_y y) (patch_z z) -> b (x y z) (patch_x patch_y patch_z c)',
                                patch_x=self.p, patch_y=self.p, patch_z=self.p)

        batch_size, tokens, _ = img_patches.shape

        # project patches with linear layer + add pos emb
        img_patches = self.project_patches(img_patches)

        # add pos. embeddings. + dropout
        # indexing with the current batch's token length to support variable sequences
        img_patches = img_patches + self.pos_emb1D
        patch_embeddings = self.emb_dropout(img_patches)

        # feed patch_embeddings and output of transformer. shape: [batch, tokens, dim]
        y = self.transformer(patch_embeddings, mask)
        # take mid token only and take final segmentation for self.n ** 3 sub-sub-volume :)
        y = self.mlp_seg_head(y[:, self.mid_token, :])

        return rearrange(y,'b (x y z classes) -> b classes x y z',x=self.n, y=self.n, z=self.n )



import torch
import torch.nn as nn

from .linformer import LinformerAttention
from ..transformer_vanilla import TransformerBlock


class LinformerBlock(TransformerBlock):
    def __init__(self, dim, heads=8, dim_head=None,
                 dim_linear_block=1024, dropout=0.1,
                 shared_projection=False, proj_shape=None,
                 trainable_proj=False, activation=nn.GELU):
        super().__init__(dim=dim, dim_linear_block=dim_linear_block, dropout=dropout, activation=activation)
        self.mhsa = LinformerAttention(dim=dim,
                                       heads=heads,
                                       dim_head=dim_head,
                                       shared_projection=shared_projection,
                                       proj_shape=proj_shape,
                                       trainable_proj=trainable_proj)

    def forward(self, x, proj_mat=None):
        return super().forward(x, proj_mat)


class LinformerEncoder(nn.Module):
    def __init__(self, dim, tokens, k=None, blocks=4, heads=8, dim_head=None,
                 dim_linear_block=1024, dropout=0.1,
                 shared_projection=True,
                 trainable_proj=True, activation=nn.GELU):
        """
        Based on the Linformer paper
        Link: https://arxiv.org/pdf/2006.04768.pdf

        Args:
            dim: token's dimension, i.e. word embedding vector size
            tokens: sequence length
            blocks: number of sequential blocks
            heads: the number of distinct representations to learn per block
            dim_head: the dim of the head.
            dim_linear_block: reprojection dim. usually multiple of dim
            dropout: dropout in mhsa block
            activation: MHSA block activation

            Specific parameters for Linformer:

            Headwise sharing and Key-value sharing are on by default!

            shared_projection: if the projection matrix E will be shared among layers
            (it will have to be passed in the forward that way)
            trainable_proj: you can enable this option to make it trainable (non-trainable in the paper)
            tokens: (tokens, projection dimension k), where k is the projection dimension of the linformer

            Choice of k for sequences of length n so that
            Linformer’s performance is nearly on par with the original Transformer:
            Practical:
            k = 128 for n = 512
            k = 256 for n = 1024
            Default is n/4 as in most of the paper's experiments
            """
        super().__init__()
        self.shared_projection = shared_projection
        self.k = k if k is not None else tokens // 4
        proj_shape = [tokens, self.k]

        if self.shared_projection:
            self.E = torch.nn.Parameter(torch.randn(proj_shape), requires_grad=trainable_proj)

        self.block_list = [LinformerBlock(dim=dim, heads=heads, dim_head=dim_head,
                                          dim_linear_block=dim_linear_block, dropout=dropout,
                                          shared_projection=shared_projection, proj_shape=proj_shape,
                                          trainable_proj=trainable_proj, activation=activation)
                           for _ in range(blocks)]
        self.layers = nn.ModuleList(self.block_list)

    def forward(self, x):
        for layer in self.layers:
            if self.shared_projection:
                x = layer(x, self.E)
            else:
                x = layer(x)
        return x



import torch.nn as nn
from einops import rearrange

from self_attention_cv.pos_embeddings.relative_embeddings_1D import RelPosEmb1D


class RelPosEmb2D(nn.Module):
    def __init__(self, feat_map_size, dim_head, heads=None):
        """
        Based on Bottleneck transformer paper
        paper: https://arxiv.org/abs/2101.11605 . Figure 4
        Output: qr^T [batch head tokens tokens]
        Args:
            tokens: the number of the tokens of the seq
            dim_head: the size of the last dimension of q

            heads: if None representation is shared across heads.
            else the number of heads must be provided
        """
        super().__init__()
        self.h, self.w = feat_map_size  # height , width
        self.total_tokens = self.h * self.w
        self.shared_heads = heads if heads is not None else True

        self.emb_w = RelPosEmb1D(self.h, dim_head, heads)
        self.emb_h = RelPosEmb1D(self.w, dim_head, heads)

    def expand_emb(self, r, dim_size):
        # Decompose and unsqueeze dimension
        r = rearrange(r, 'b (h x) i j -> b h x () i j', x=dim_size)
        expand_index = [-1, -1, -1, dim_size, -1, -1]  # -1 indicates no expansion
        r = r.expand(expand_index)
        return rearrange(r, 'b h x1 x2 y1 y2 -> b h (x1 y1) (x2 y2)')

    def forward(self, q):
        """
        Args:
            q: [batch, heads, tokens, dim_head]
        Returns: [ batch, heads, tokens, tokens]
        """
        assert self.total_tokens == q.shape[2], f'Tokens {q.shape[2]} of q must \
        be equal to the product of the feat map size {self.total_tokens} '

        # out: [batch head*w h h]
        r_h = self.emb_w(rearrange(q, 'b h (x y) d -> b (h x) y d', x=self.h, y=self.w))
        r_w = self.emb_h(rearrange(q, 'b h (x y) d -> b (h y) x d', x=self.h, y=self.w))
        q_r = self.expand_emb(r_h, self.h) + self.expand_emb(r_w, self.h)
        return q_r  # q_r transpose in figure 4 of the paper



import numpy as np
import torch
from einops import rearrange
from torch import nn


def compute_mhsa(q, k, v, scale_factor=1, mask=None):
    # resulted shape will be: [batch, heads, tokens, tokens]
    scaled_dot_prod = torch.einsum('... i d , ... j d -> ... i j', q, k) * scale_factor

    if mask is not None:
        assert mask.shape == scaled_dot_prod.shape[2:]
        scaled_dot_prod = scaled_dot_prod.masked_fill(mask, -np.inf)

    attention = torch.softmax(scaled_dot_prod, dim=-1)
    # calc result per head
    return torch.einsum('... i j , ... j d -> ... i d', attention, v)


class MultiHeadSelfAttention(nn.Module):
    def __init__(self, dim, heads=8, dim_head=None):
        """
        Implementation of multi-head attention layer of the original transformer model.
        einsum and einops.rearrange is used whenever possible
        Args:
            dim: token's dimension, i.e. word embedding vector size
            heads: the number of distinct representations to learn
            dim_head: the dim of the head. In general dim_head<dim.
            However, it may not necessary be (dim/heads)
        """
        super().__init__()
        self.dim_head = (int(dim / heads)) if dim_head is None else dim_head
        _dim = self.dim_head * heads
        self.heads = heads
        self.to_qvk = nn.Linear(dim, _dim * 3, bias=False)
        self.W_0 = nn.Linear(_dim, dim, bias=False)
        self.scale_factor = self.dim_head ** -0.5

    def forward(self, x, mask=None):
        assert x.dim() == 3
        qkv = self.to_qvk(x)  # [batch, tokens, dim*3*heads ]

        # decomposition to q,v,k and cast to tuple
        # the resulted shape before casting to tuple will be: [3, batch, heads, tokens, dim_head]
        q, k, v = tuple(rearrange(qkv, 'b t (d k h ) -> k b h t d ', k=3, h=self.heads))

        out = compute_mhsa(q, k, v, mask=mask, scale_factor=self.scale_factor)

        # re-compose: merge heads with dim_head
        out = rearrange(out, "b h t d -> b t (h d)")
        # Apply final linear transformation layer
        return self.W_0(out)



from einops import rearrange
from torch import nn

from .tied_axial_attention import TiedRowAxialAttention
from ..transformer_vanilla.mhsa import MultiHeadSelfAttention


class MSATransformerBlock(nn.Module):
    """
    MSA transformer block from the paper MSA Transformer
    Link: https://www.biorxiv.org/content/10.1101/2021.02.12.430858v1
    """

    def __init__(self, *, dim, rows, heads=8, dim_head=None, dim_linear_block=1024, dropout=0.1, activation=nn.GELU):
        """
        Args:
            dim: token's vector length
            rows: number of rows with shared/tied attention that will be summed in Q*K^T
            heads: number of heads
            dim_head: if none dim/heads is used
            dim_linear_block: the inner projection dim
            dropout: probability of droppping values
        """
        super().__init__()
        self.column_att = MultiHeadSelfAttention(dim=dim, heads=heads, dim_head=dim_head)
        self.row_att = TiedRowAxialAttention(dim=dim, rows=rows, heads=heads, dim_head=dim_head)
        self.rows = rows
        self.dim = dim

        self.norm_1 = nn.LayerNorm(dim)
        self.norm_2 = nn.LayerNorm(dim)

        # as in vanilla transformer
        self.mlp = nn.Sequential(
            nn.LayerNorm(dim),
            nn.Linear(dim, dim_linear_block),
            activation(),  # nn.ReLU or nn.GELU
            nn.Dropout(dropout),
            nn.Linear(dim_linear_block, dim),
            nn.Dropout(dropout)
        )

    def forward(self, x, mask=None):
        assert x.dim() == 4, 'Ensure your input is 4D: [batch,channels, height,width]'
        # compose rows (h) with batch
        x = rearrange(x, 'b c h w -> (b h) c w')
        x = self.row_att(self.norm_1(x), mask) + x

        # decompose rows + merge batch with height
        x = rearrange(x, '(b h) c w  -> (b w) c h', h=self.rows)
        x = self.column_att(self.norm_2(x), mask) + x
        x = self.mlp(x) + x
        return rearrange(x, '(b w) c h -> b c h w', w=self.dim)


class MSATransformerEncoder(nn.Module):
    def __init__(self, *, dim, rows, blocks=6, heads=8, dim_head=None, dim_linear_block=1024, dropout=0):
        super().__init__()
        self.block_list = [MSATransformerBlock(dim=dim, rows=rows,
                                               heads=heads,
                                               dim_head=dim_head,
                                               dim_linear_block=dim_linear_block,
                                               dropout=dropout) for _ in range(blocks)]
        self.layers = nn.ModuleList(self.block_list)

    def forward(self, x, mask=None):
        for layer in self.layers:
            x = layer(x, mask)
        return x



import torch
from einops import rearrange
from torch import nn, einsum

from ..pos_embeddings import RelPosEmb2D


class BottleneckAttention(nn.Module):
    def __init__(
            self,
            dim,
            fmap_size,
            heads=4,
            dim_head=None,
            content_positional_embedding=True
    ):
        """
        tensorflow code gist
        https://gist.github.com/aravindsrinivas/56359b79f0ce4449bcb04ab4b56a57a2
        lucidrains' code that I also studied until I finally understood how it works:
        https://github.com/lucidrains/bottleneck-transformer-pytorch
        paper: https://arxiv.org/abs/2101.11605
        Args:
            dim: dimension of the token vector (d_model)
            fmap_size: tuple with the feat map spatial dims
            heads: number of heads representations
            dim_head: inner dimension of the head. dim / heads by default
            content_positional_embedding: whether to include the 2 rel. pos embedding for the query
        """
        super().__init__()
        self.heads = heads
        self.dim_head = (int(dim / heads)) if dim_head is None else dim_head
        self.scale = self.dim_head ** -0.5
        self.fmap_size = fmap_size
        self.content_positional_embedding = content_positional_embedding

        self.to_qkv = nn.Conv2d(dim, heads * self.dim_head * 3, 1, bias=False)

        self.height = self.fmap_size[0]
        self.width = self.fmap_size[1]

        if self.content_positional_embedding:
            self.pos_emb2D = RelPosEmb2D(feat_map_size=fmap_size, dim_head=self.dim_head)

    def forward(self, x):
        assert x.dim() == 4, f'Expected 4D tensor, got {x.dim()}D tensor'

        # [batch (heads*3*dim_head) height width]
        qkv = self.to_qkv(x)
        # decompose heads and merge spatial dims as tokens
        q, k, v = tuple(rearrange(qkv, 'b (d k h ) x y  -> k b h (x y) d', k=3, h=self.heads))

        # i, j refer to tokens
        dot_prod = einsum('b h i d, b h j d -> b h i j', q, k) * self.scale

        if self.content_positional_embedding:
            dot_prod = dot_prod + self.pos_emb2D(q)

        attention = torch.softmax(dot_prod, dim=-1)
        out = einsum('b h i j, b h j d -> b h i d', attention, v)
        # Merge heads and decompose tokens to spatial dims
        out = rearrange(out, 'b h (x y) d -> b (h d) x y', x=self.height, y=self.width)
        return out



import os
import random
from typing import List, Tuple

import numpy as np
import torch
from einops import repeat
from torch import Tensor, nn


def expand_to_batch(tensor, desired_size):
    tile = desired_size // tensor.shape[0]
    return repeat(tensor, 'b ... -> (b tile) ...', tile=tile)


def init_random_seed(seed, gpu=False):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    if gpu:
        torch.backends.cudnn.deterministic = True


# from https://huggingface.co/transformers/_modules/transformers/modeling_utils.html
def get_module_device(parameter: nn.Module):
    try:
        return next(parameter.parameters()).device
    except StopIteration:
        # For nn.DataParallel compatibility in PyTorch 1.5

        def find_tensor_attributes(module: nn.Module) -> List[Tuple[str, Tensor]]:
            tuples = [(k, v) for k, v in module.__dict__.items() if torch.is_tensor(v)]
            return tuples

        gen = parameter._named_members(get_members_fn=find_tensor_attributes)
        first_tuple = next(gen)
        return first_tuple[1].device



import torch
import torch.nn as nn
from einops import rearrange

from self_attention_cv.UnetTr.modules import TranspConv3DBlock, BlueBlock, Conv3DBlock
from self_attention_cv.UnetTr.volume_embedding import Embeddings3D
from self_attention_cv.transformer_vanilla import TransformerBlock


class TransformerEncoder(nn.Module):
    def __init__(self, embed_dim, num_heads, num_layers, dropout, extract_layers):
        super().__init__()
        self.layer = nn.ModuleList()
        self.extract_layers = extract_layers

        # makes TransformerBlock device aware
        self.block_list = nn.ModuleList()
        for _ in range(num_layers):
            self.block_list.append(TransformerBlock(dim=embed_dim, heads=num_heads,
                                            dim_linear_block=1024, dropout=dropout, prenorm=True))

    def forward(self, x):
        extract_layers = []
        for depth, layer_block in enumerate(self.block_list):
            x = layer_block(x)
            if (depth + 1) in self.extract_layers:
                extract_layers.append(x)

        return extract_layers

# based on https://arxiv.org/abs/2103.10504
class UNETR(nn.Module):
    def __init__(self, img_shape=(128, 128, 128), input_dim=4, output_dim=3,
                 embed_dim=768, patch_size=16, num_heads=12, dropout=0.1,
                 num_layers=12, ext_layers=[3, 6, 9, 12], version='light'):
        """

        Args:
            img_shape: volume shape, provided as a tuple
            input_dim: input modalities/channels
            output_dim: number of classes
            embed_dim: transformer embed dim.
            patch_size: the non-overlapping patches to be created
            num_heads: for the transformer encoder
            dropout: percentage for dropout
            num_layers: static to the architecture. cannot be changed with the current architecture.
            ext_layers: transformer layers to use their output
            version: 'light' saves some parameters in the decoding part
        """
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.embed_dim = embed_dim
        self.img_shape = img_shape
        self.patch_size = patch_size
        self.num_heads = num_heads
        self.dropout = dropout
        self.num_layers = num_layers
        self.ext_layers = ext_layers
        self.patch_dim = [int(x / patch_size) for x in img_shape]
        self.base_filters = 64
        self.prelast_filters = 32

        # cheap way to reduce the number of parameters in the decoding part.
        self.yellow_conv_channels = [256, 128, 64] if version == 'light' else [512, 256, 128]

        self.embed = Embeddings3D(input_dim=input_dim, embed_dim=embed_dim,
                                  cube_size=img_shape, patch_size=patch_size, dropout=dropout)

        self.transformer = TransformerEncoder(embed_dim, num_heads, num_layers, dropout, ext_layers)

        self.init_conv = Conv3DBlock(input_dim, self.base_filters, double=True)

        # blue blocks in Fig.1
        self.z3_blue_conv = nn.Sequential(
            BlueBlock(in_planes=embed_dim, out_planes=512),
            BlueBlock(in_planes=512, out_planes=256),
            BlueBlock(in_planes=256, out_planes=128))

        self.z6_blue_conv = nn.Sequential(
            BlueBlock(in_planes=embed_dim, out_planes=512),
            BlueBlock(in_planes=512, out_planes=256))

        self.z9_blue_conv = BlueBlock(in_planes=embed_dim, out_planes=512)

        # Green blocks in Fig.1
        self.z12_deconv = TranspConv3DBlock(embed_dim, 512)

        self.z9_deconv = TranspConv3DBlock(self.yellow_conv_channels[0], 256)
        self.z6_deconv = TranspConv3DBlock(self.yellow_conv_channels[1], 128)
        self.z3_deconv = TranspConv3DBlock(self.yellow_conv_channels[2], 64)

        # Yellow blocks in Fig.1
        self.z9_conv = Conv3DBlock(1024, self.yellow_conv_channels[0], double=True)
        self.z6_conv = Conv3DBlock(512, self.yellow_conv_channels[1], double=True)
        self.z3_conv = Conv3DBlock(256, self.yellow_conv_channels[2], double=True)
        # out convolutions
        self.out_conv = nn.Sequential(
            # last yellow conv block
            Conv3DBlock(128, self.prelast_filters, double=True),
            # grey block, final classification layer
            Conv3DBlock(self.prelast_filters, output_dim, kernel_size=1, double=False))

    def forward(self, x):
        transf_input = self.embed(x)
        z3, z6, z9, z12 = map(lambda t: rearrange(t, 'b (x y z) d -> b d x y z',
                                                  x=self.patch_dim[0], y=self.patch_dim[1], z=self.patch_dim[2]),
                              self.transformer(transf_input))

        # Blue convs
        z0 = self.init_conv(x)
        z3 = self.z3_blue_conv(z3)
        z6 = self.z6_blue_conv(z6)
        z9 = self.z9_blue_conv(z9)

        # Green block for z12
        z12 = self.z12_deconv(z12)
        # Concat + yellow conv
        y = torch.cat([z12, z9], dim=1)
        y = self.z9_conv(y)

        # Green block for z6
        y = self.z9_deconv(y)
        # Concat + yellow conv
        y = torch.cat([y, z6], dim=1)
        y = self.z6_conv(y)

        # Green block for z3
        y = self.z6_deconv(y)
        # Concat + yellow conv
        y = torch.cat([y, z3], dim=1)
        y = self.z3_conv(y)

        y = self.z3_deconv(y)
        y = torch.cat([y, z0], dim=1)
        return self.out_conv(y)



import numpy as np
import torch
from einops import rearrange
from torch import nn


class TiedRowAxialAttention(nn.Module):
    def __init__(self, *, dim, rows, heads=8, dim_head=None):
        """
        Tied row attention uses a single attention map for all tokens in the MSA
        Applies tied attention by decomposing batches*rows and summing Q*K^T
        over all rows. batches*rows is recomposed by multipling the attention weights back
        with the value vector
        The  Equation 1 in the paper.
        Tied attention reduces the memory footprint of the row attentions
        from O(rows dim^2) to O(dim^2).
        Link: https://www.biorxiv.org/content/10.1101/2021.02.12.430858v1
        Args:
            dim: token's dimension i.e column pixels of an image
            rows: number of rows with shared/tied attention that will be summed in Q*K^T
            heads: the number of distinct representations to learn
            dim_head: the dim of the head.
        """
        super().__init__()
        self.dim_head = (int(dim / heads)) if dim_head is None else dim_head
        _dim = self.dim_head * heads
        self.heads = heads
        self.to_qvk = nn.Linear(dim, _dim * 3, bias=False)
        self.W_0 = nn.Linear(_dim, dim, bias=False)
        self.rows = rows
        self.scale_factor = (self.rows * self.dim_head) ** -0.5

    def forward(self, x, mask=None):
        assert x.dim() == 3, 'Ensure your input is 4D: [b * width, chan, height] or [b * height, chan, width]'
        qkv = self.to_qvk(x)  # [batch, tokens, dim*3*heads ]

        # decompose heads, (q,v,k) tuple , and rows
        # the resulted shape before casting to tuple will be: [3, batch, heads, tokens, dim_head]
        q, k, v = tuple(rearrange(qkv, '(b rows) t (d k h ) -> k b rows h t d ', k=3, h=self.heads, rows=self.rows))

        # resulted shape will be: [batch, heads, tokens, tokens]. Notice that the row index is vanished
        # meaning that this dimension is summed
        scaled_dot_prod = torch.einsum('b r h i d , b r h j d -> b h i j', q, k) * self.scale_factor

        if mask is not None:
            assert mask.shape == scaled_dot_prod.shape[2:]
            scaled_dot_prod = scaled_dot_prod.masked_fill(mask, -np.inf)

        attention = torch.softmax(scaled_dot_prod, dim=-1)
        # calc result per head and row for v
        out = torch.einsum('b h i j , b r h j d -> b r h i d', attention, v)
        # re-compose: merge heads with dim_head
        out = rearrange(out, "b rows h t d -> (b rows) t (h d)")
        # Apply final linear transformation layer
        return self.W_0(out)



from einops import rearrange
from torch import nn

from .axial_attention import AxialAttention
from ..transformer_vanilla.mhsa import MultiHeadSelfAttention


def _conv2d1x1(in_channels, out_channels, stride=1):
    """1x1 convolution for contraction and expansion of the channels dimension
    conv is followed by batch norm"""
    return nn.Sequential(nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False),
                         nn.BatchNorm2d(out_channels))


class AxialAttentionBlock(nn.Module):
    def __init__(self, in_channels, dim, heads=8, axial_att=True, dim_head=None):
        """
        Axial-attention block implementation as described in:
        paper: https://arxiv.org/abs/2003.07853 , Fig. 2 page 7
        blogpost: TBA
        official code: https://github.com/csrhddlam/axial-deeplab
        Args:
            in_channels:
            dim: token's dim
            heads: the number of distict head representations
            axial_att: whether to use axial att or MHSA
            dim_head: for MHSA only
        """
        super().__init__()
        self.dim = dim
        self.heads = heads
        d_in = 128  # hardcoded

        # brings the input channels to 128 feature maps
        self.in_conv1x1 = _conv2d1x1(in_channels, d_in)
        self.out_conv1x1 = _conv2d1x1(d_in, in_channels)
        self.relu = nn.ReLU(inplace=True)

        if axial_att:
            self.dim_head = d_in // self.heads
            self.height_att = AxialAttention(dim=dim, in_channels=d_in, heads=heads)
            self.width_att = AxialAttention(dim=dim, in_channels=d_in, heads=heads)
        else:
            self.dim_head = (int(dim / heads)) if dim_head is None else dim_head
            self.width_att = MultiHeadSelfAttention(dim=dim, dim_head=dim_head, heads=heads)
            self.height_att = MultiHeadSelfAttention(dim=dim, dim_head=dim_head, heads=heads)

    def forward(self, x_in):
        assert x_in.dim() == 4, 'Ensure your input is 4D: [batch,channels, height,width]'
        x = self.relu(self.in_conv1x1(x_in))
        # merge batch dim with width
        x = rearrange(x, 'b c h w -> (b w) c h')
        x = self.height_att(x)
        # decompose width + merge batch with height
        x = rearrange(x, '(b w) c h  -> (b h) c w', w=self.dim)
        x = self.relu(self.width_att(x))
        x = rearrange(x, '(b h) c w -> b c h w', h=self.dim)
        return self.relu(self.out_conv1x1(x) + x_in)



import numpy as np
import torch
from einops import rearrange
from torch import nn


class SelfAttention(nn.Module):
    """
    Implementation of plain self attention mechanism with einsum operations
    Paper: https://arxiv.org/abs/1706.03762
    Blog: https://theaisummer.com/transformer/
    """

    def __init__(self, dim):
        """
        Args:
            dim: for NLP it is the dimension of the embedding vector
            the last dimension size that will be provided in forward(x)
            where x is a 3D tensor
        """
        super().__init__()
        self.to_qvk = nn.Linear(dim, dim * 3, bias=False)
        self.scale_factor = dim ** -0.5  # 1/np.sqrt(dim)

    def forward(self, x, mask=None):
        assert x.dim() == 3, '3D tensor must be provided'
        qkv = self.to_qvk(x)  # [batch, tokens, dim*3 ]

        # decomposition to q,v,k
        # rearrange tensor to [3, batch, tokens, dim] and cast to tuple
        q, k, v = tuple(rearrange(qkv, 'b t (d k) -> k b t d ', k=3))

        # Resulting shape: [batch, tokens, tokens]
        scaled_dot_prod = torch.einsum('b i d , b j d -> b i j', q, k) * self.scale_factor

        if mask is not None:
            assert mask.shape == scaled_dot_prod.shape[1:]
            scaled_dot_prod = scaled_dot_prod.masked_fill(mask, -np.inf)

        attention = torch.softmax(scaled_dot_prod, dim=-1)
        return torch.einsum('b i j , b j d -> b i d', attention, v)



import torch
import torch.nn as nn
from einops import rearrange

from .spacetime_attention import SpacetimeMHSA
from ..common import expand_to_batch

class TimeSformerBlock(nn.Module):
    def __init__(self, *, frames, patches, dim=512,
                 heads=None, dim_linear_block=1024,
                 activation=nn.GELU,
                 dropout=0.1, classification=True,
                 linear_spatial_attention=False, k=None):
        """
        Args:
            dim: token's dim
            heads: number of heads
            linear_spatial_attention: if True Linformer-based attention is applied
        """
        super().__init__()
        self.frames = frames
        self.patches = patches
        self.classification = classification

        self.time_att = nn.Sequential(nn.LayerNorm(dim),
                                      SpacetimeMHSA(dim, tokens_to_attend=self.frames, space_att=False,
                                                    heads=heads, classification=self.classification))

        self.space_att = nn.Sequential(nn.LayerNorm(dim),
                                       SpacetimeMHSA(dim, tokens_to_attend=self.patches, space_att=True,
                                                     heads=heads, classification=self.classification,
                                                     linear_spatial_attention=linear_spatial_attention, k=k))

        self.mlp = nn.Sequential(
            nn.LayerNorm(dim),
            nn.Linear(dim, dim_linear_block),
            activation(),
            nn.Dropout(dropout),
            nn.Linear(dim_linear_block, dim),
            nn.Dropout(dropout)
        )

    def forward(self, x):
        assert x.dim() == 3
        x = self.time_att(x) + x
        x = self.space_att(x) + x
        x = self.mlp(x) + x
        return x


class Timesformer(nn.Module):
    def __init__(self, *,
                 img_dim, frames,
                 num_classes=None,
                 in_channels=3,
                 patch_dim=16,
                 dim=512,
                 blocks=6,
                 heads=4,
                 dim_linear_block=1024,
                 dim_head=None,
                 activation=nn.GELU,
                 dropout=0,
                 linear_spatial_attention=False, k=None):
        """
        Adapting ViT for video classification.
        Best strategy to handle multiple frames so far is
        Divided Space-Time Attention (T+S). We apply attention to projected
        image patches, first in time and then in both spatial dims.
        Args:
            img_dim: the spatial image size
            frames: video frames
            num_classes: classification task classes
            in_channels: number of img channels
            patch_dim: desired patch dim

            dim: the linear layer's dim to project the patches for MHSA
            blocks: number of transformer blocks
            heads: number of heads
            dim_linear_block: inner dim of the transformer linear block
            dim_head: dim head in case you want to define it. defaults to dim/heads
            dropout: for pos emb and transformer
            linear_spatial_attention: for Linformer linear attention
            k: for Linformer linear attention
        """
        super().__init__()
        assert img_dim % patch_dim == 0, f'patch size {patch_dim} not divisible by img dim {img_dim}'
        self.p = patch_dim
        # classification: creates an extra CLS token that we will index in the final classification layer
        self.classification = True if num_classes is not None else False
        img_patches = (img_dim // patch_dim) ** 2
        # tokens = number of img patches * number of frames
        tokens_spacetime = frames * img_patches
        self.token_dim = in_channels * (patch_dim ** 2)
        self.dim = dim
        self.dim_head = (int(dim / heads)) if dim_head is None else dim_head

        # Projection and pos embeddings
        self.project_patches = nn.Linear(self.token_dim, dim)

        self.emb_dropout = nn.Dropout(dropout)

        self.cls_token = nn.Parameter(torch.randn(1, 1, dim))
        self.pos_emb1D = nn.Parameter(torch.randn(tokens_spacetime + 1, dim))

        self.mlp_head = nn.Sequential(nn.LayerNorm(dim), nn.Linear(dim, num_classes))

        transormer_blocks = [TimeSformerBlock(
            frames=frames, patches=img_patches, dim=dim,
            heads=heads, dim_linear_block=dim_linear_block,
            activation=activation,
            dropout=dropout,
            linear_spatial_attention=linear_spatial_attention, k=k)
            for _ in range(blocks)]

        self.transformer = nn.Sequential(*transormer_blocks)

    def forward(self, vid):
        # Create patches as in ViT wherein frames are merged with patches
        # from [batch, frames, channels, h, w] to [batch, tokens , N], N=p*p*c , tokens = h/p *w/p
        img_patches = rearrange(vid,
                                'b f c (patch_x x) (patch_y y) -> b (f x y) (patch_x patch_y c)',
                                patch_x=self.p, patch_y=self.p)

        batch_size, tokens_spacetime, _ = img_patches.shape

        # project patches with linear layer + add pos emb
        img_patches = self.project_patches(img_patches)

        img_patches = torch.cat((expand_to_batch(self.cls_token, batch_size), img_patches), dim=1)
        # add pos. embeddings. + dropout
        # indexing with the current batch's token length to support variable sequences
        img_patches = img_patches + self.pos_emb1D[:tokens_spacetime + 1, :]
        patch_embeddings = self.emb_dropout(img_patches)

        # feed patch_embeddings and output of transformer. shape: [batch, tokens, dim]
        y = self.transformer(patch_embeddings)

        # we index only the cls token for classification. nlp tricks :P
        return self.mlp_head(y[:, 0, :]) if self.classification else y



import torch
from einops import rearrange
from torch import nn


class Relative2DPosEncQKV(nn.Module):
    def __init__(self, dim_head, dim_v=16, dim_kq=8):
        """
        Implementation of 2D relative positional embeddings for q,v,k
        Out shape shape will be [dim_head, dim, dim]
        Embeddings are shared across heads for all q,k,v
        Based on Axial DeepLab https://arxiv.org/abs/2003.07853

        Args:
            dim_head: the dimension of the head
            dim_v: d_out in the paper
            dim_kq: d_k in the paper
        """
        super().__init__()
        self.dim = dim_head
        self.dim_head_v = dim_v
        self.dim_head_kq = dim_kq
        self.qkv_chan = 2 * self.dim_head_kq + self.dim_head_v

        # 2D relative position embeddings of q,k,v:
        self.relative = nn.Parameter(torch.randn(self.qkv_chan, dim_head * 2 - 1), requires_grad=True)
        self.relative_index_2d = self.relative_index()

    def relative_index(self):
        # integer lists from 0 to 63
        query_index = torch.arange(self.dim).unsqueeze(0)  # [1, dim]
        key_index = torch.arange(self.dim).unsqueeze(1)  # [dim, 1]

        relative_index_2d = (key_index - query_index) + self.dim - 1  # dim X dim
        return rearrange(relative_index_2d, 'i j->(i j)')  # flatten

    def forward(self):
        rel_indx = self.relative_index_2d.to(self.relative.device)
        all_embeddings = torch.index_select(self.relative, 1, rel_indx)  # [head_planes , (dim*dim)]

        all_embeddings = rearrange(all_embeddings, ' c (x y)  -> c x y', x=self.dim)

        q_embedding, k_embedding, v_embedding = torch.split(all_embeddings,
                                                            [self.dim_head_kq, self.dim_head_kq, self.dim_head_v],
                                                            dim=0)
        return q_embedding, k_embedding, v_embedding



import torch
from einops import rearrange
from torch import nn

from ..transformer_vanilla.mhsa import compute_mhsa


def project_vk_linformer(v, k, E):
    # project k,v
    v = torch.einsum('b h j d , j k -> b h k d', v, E)
    k = torch.einsum('b h j d , j k -> b h k d', k, E)
    return v, k


class LinformerAttention(nn.Module):
    def __init__(self, dim, heads=8, dim_head=None, shared_projection=True, proj_shape=None, trainable_proj=True):
        """
        Based on the Linformer paper
        Link: https://arxiv.org/pdf/2006.04768.pdf

        Args:
            dim: token's dimension, i.e. word embedding vector size
            heads: the number of distinct representations to learn
            dim_head: the dim of the head.

            shared_projection: if the projection matrix will be shared among layers
            (it will have to be passed in the forward that way)
            trainable_proj: if the projection matrix E matrix is not shared,
            you can enable this option to make it trainable (non trainable in the paper)
            proj_shape: 2-tuple (tokens,k), where k is the projection dimension of the linformer
            """
        super().__init__()
        self.dim_head = (int(dim / heads)) if dim_head is None else dim_head
        _dim = self.dim_head * heads
        self.heads = heads
        self.to_qvk = nn.Linear(dim, _dim * 3, bias=False)
        self.W_0 = nn.Linear(_dim, dim, bias=False)
        self.scale_factor = self.dim_head ** -0.5
        self.shared_projection = shared_projection

        if not shared_projection:
            self.E = torch.nn.Parameter(torch.randn(proj_shape), requires_grad=trainable_proj)
            self.k = proj_shape[1]

    def forward(self, x, proj_mat=None):
        assert x.dim() == 3
        E = proj_mat if (self.shared_projection and proj_mat is not None) else self.E
        assert x.shape[1] == E.shape[0], f'{x.shape[1]} Token in the input sequence while' \
                                         f' {E.shape[0]} were provided in the E proj matrix'

        qkv = self.to_qvk(x)  # [batch, tokens, dim*3*heads ]
        q, k, v = tuple(rearrange(qkv, 'b t (d k h ) -> k b h t d ', k=3, h=self.heads))

        v, k = project_vk_linformer(v, k, E)

        out = compute_mhsa(q, k, v, scale_factor=self.scale_factor)
        # re-compose: merge heads with dim_head

        out = rearrange(out, "b h i d -> b i (h d)")
        # Apply final linear transformation layer
        return self.W_0(out)



import torch
import torch.nn as nn
from ..common import expand_to_batch


class AbsPositionalEncoding1D(nn.Module):
    def __init__(self, tokens, dim):
        super(AbsPositionalEncoding1D, self).__init__()
        self.abs_pos_enc = nn.Parameter(torch.randn(1,tokens, dim))

    def forward(self, x):
        batch = x.size()[0]
        return x + expand_to_batch(self.abs_pos_enc, desired_size=batch)


import torch
import torch.nn as nn


class SignleConv(nn.Module):
    """
    Double convolution block that keeps that spatial sizes the same
    """

    def __init__(self, in_ch, out_ch, norm_layer=None):
        super(SignleConv, self).__init__()

        if norm_layer is None:
            norm_layer = nn.BatchNorm2d

        self.conv = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, kernel_size=3, stride=1, padding=1),
            norm_layer(out_ch),
            nn.ReLU(inplace=True))

    def forward(self, x):
        return self.conv(x)


class DoubleConv(nn.Module):
    """
    Double convolution block that keeps that spatial sizes the same
    """

    def __init__(self, in_ch, out_ch, norm_layer=None):
        super(DoubleConv, self).__init__()
        self.conv = nn.Sequential(SignleConv(in_ch, out_ch, norm_layer),
                                  SignleConv(out_ch, out_ch, norm_layer))

    def forward(self, x):
        return self.conv(x)


class Up(nn.Module):
    """
    Doubles spatial size with bilinear upsampling
    Skip connections and double convs
    """

    def __init__(self, in_ch, out_ch):
        super(Up, self).__init__()
        mode = "bilinear"
        self.up = nn.Upsample(scale_factor=2, mode=mode, align_corners=True)
        self.conv = DoubleConv(in_ch, out_ch)

    def forward(self, x1, x2=None):
        """
        Args:
            x1: [b,c, h, w]
            x2: [b,c, 2*h,2*w]

        Returns: 2x upsampled double conv reselt
        """
        x = self.up(x1)
        if x2 is not None:
            x = torch.cat([x2, x], dim=1)
        return self.conv(x)



import torch
import torch.nn as nn
from einops import rearrange


# borrowed from
# https://github.com/lucidrains/bottleneck-transformer-pytorch/blob/main/bottleneck_transformer_pytorch/bottleneck_transformer_pytorch.py#L21
# i will try to reimplement the function
# as soon as i understand how it works
# not clear to me how it works yet
def relative_to_absolute(q):
    """
    Converts the dimension that is specified from the axis
    from relative distances (with length 2*tokens-1) to absolute distance (length tokens)
      Input: [bs, heads, length, 2*length - 1]
      Output: [bs, heads, length, length]
    """
    b, h, l, _, device, dtype = *q.shape, q.device, q.dtype
    dd = {'device': device, 'dtype': dtype}
    col_pad = torch.zeros((b, h, l, 1), **dd)
    x = torch.cat((q, col_pad), dim=3)  # zero pad 2l-1 to 2l
    flat_x = rearrange(x, 'b h l c -> b h (l c)')
    flat_pad = torch.zeros((b, h, l - 1), **dd)
    flat_x_padded = torch.cat((flat_x, flat_pad), dim=2)
    final_x = flat_x_padded.reshape(b, h, l + 1, 2 * l - 1)
    final_x = final_x[:, :, :l, (l - 1):]
    return final_x


def rel_pos_emb_1d(q, rel_emb, shared_heads):
    """
    Same functionality as RelPosEmb1D

    Args:
        q: a 4d tensor of shape [batch, heads, tokens, dim]
        rel_emb: a 2D or 3D tensor
        of shape [ 2*tokens-1 , dim] or [ heads, 2*tokens-1 , dim]
    """
    if shared_heads:
        emb = torch.einsum('b h t d, r d -> b h t r', q, rel_emb)
    else:
        emb = torch.einsum('b h t d, h r d -> b h t r', q, rel_emb)
    return relative_to_absolute(emb)


class RelPosEmb1D(nn.Module):
    def __init__(self, tokens, dim_head, heads=None):
        """
        Output: [batch head tokens tokens]
        Args:
            tokens: the number of the tokens of the seq
            dim_head: the size of the last dimension of q

            heads: if None representation is shared across heads.
            else the number of heads must be provided
        """
        super().__init__()
        scale = dim_head ** -0.5
        self.shared_heads = heads if heads is not None else True
        if self.shared_heads:
            self.rel_pos_emb = nn.Parameter(torch.randn(2 * tokens - 1, dim_head) * scale)
        else:
            self.rel_pos_emb = nn.Parameter(torch.randn(heads, 2 * tokens - 1, dim_head) * scale)

    def forward(self, q):
        return rel_pos_emb_1d(q, self.rel_pos_emb, self.shared_heads)



import torch
import torch.nn as nn
from einops import rearrange

from self_attention_cv import TransformerEncoder
from ..common import expand_to_batch


class ViT(nn.Module):
    def __init__(self, *,
                 img_dim,
                 in_channels=3,
                 patch_dim=16,
                 num_classes=10,
                 dim=512,
                 blocks=6,
                 heads=4,
                 dim_linear_block=1024,
                 dim_head=None,
                 dropout=0, transformer=None, classification=True):
        """
        Minimal re-implementation of ViT
        Args:
            img_dim: the spatial image size
            in_channels: number of img channels
            patch_dim: desired patch dim
            num_classes: classification task classes
            dim: the linear layer's dim to project the patches for MHSA
            blocks: number of transformer blocks
            heads: number of heads
            dim_linear_block: inner dim of the transformer linear block
            dim_head: dim head in case you want to define it. defaults to dim/heads
            dropout: for pos emb and transformer
            transformer: in case you want to provide another transformer implementation
            classification: creates an extra CLS token that we will index in the final classification layer
        """
        super().__init__()
        assert img_dim % patch_dim == 0, f'patch size {patch_dim} not divisible by img dim {img_dim}'
        self.p = patch_dim
        self.classification = classification
        # tokens = number of patches
        tokens = (img_dim // patch_dim) ** 2
        self.token_dim = in_channels * (patch_dim ** 2)
        self.dim = dim
        self.dim_head = (int(dim / heads)) if dim_head is None else dim_head

        # Projection and pos embeddings
        self.project_patches = nn.Linear(self.token_dim, dim)

        self.emb_dropout = nn.Dropout(dropout)

        self.cls_token = nn.Parameter(torch.randn(1, 1, dim))
        self.pos_emb1D = nn.Parameter(torch.randn(tokens + 1, dim))
        self.mlp_head = nn.Linear(dim, num_classes)

        if transformer is None:
            self.transformer = TransformerEncoder(dim, blocks=blocks, heads=heads,
                                                  dim_head=self.dim_head,
                                                  dim_linear_block=dim_linear_block,
                                                  dropout=dropout)
        else:
            self.transformer = transformer

    def forward(self, img, mask=None):
        # Create patches
        # from [batch, channels, h, w] to [batch, tokens , N], N=p*p*c , tokens = h/p *w/p
        img_patches = rearrange(img,
                                'b c (patch_x x) (patch_y y) -> b (x y) (patch_x patch_y c)',
                                patch_x=self.p, patch_y=self.p)

        batch_size, tokens, _ = img_patches.shape

        # project patches with linear layer + add pos emb
        img_patches = self.project_patches(img_patches)

        img_patches = torch.cat((expand_to_batch(self.cls_token, desired_size=batch_size), img_patches), dim=1)

        # add pos. embeddings. + dropout
        # indexing with the current batch's token length to support variable sequences
        img_patches = img_patches + self.pos_emb1D[:tokens + 1, :]
        patch_embeddings = self.emb_dropout(img_patches)

        # feed patch_embeddings and output of transformer. shape: [batch, tokens, dim]
        y = self.transformer(patch_embeddings, mask)

        # we index only the cls token for classification. nlp tricks :P
        return self.mlp_head(y[:, 0, :]) if self.classification else y[:, 1:, :]



import torch.nn as nn


# yellow block in Fig.1
class Conv3DBlock(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size=3, double=True):
        super().__init__()
        if double:
            self.conv_block = nn.Sequential(
                nn.Conv3d(in_planes, out_planes, kernel_size=kernel_size, stride=1,
                          padding=((kernel_size - 1) // 2)),
                nn.BatchNorm3d(out_planes),
                nn.ReLU(inplace=True),
                nn.Conv3d(out_planes, out_planes, kernel_size=kernel_size, stride=1,
                          padding=((kernel_size - 1) // 2)),
                nn.BatchNorm3d(out_planes),
                nn.ReLU(inplace=True)
            )
        else:
            self.conv_block = nn.Sequential(
                nn.Conv3d(in_planes, out_planes, kernel_size=kernel_size, stride=1,
                          padding=((kernel_size - 1) // 2)),
                nn.BatchNorm3d(out_planes),
                nn.ReLU(inplace=True))

    def forward(self, x):
        return self.conv_block(x)

# green block in Fig.1
class TranspConv3DBlock(nn.Module):
    def __init__(self, in_planes, out_planes):
        super().__init__()
        self.block = nn.ConvTranspose3d(in_planes, out_planes, kernel_size=2, stride=2, padding=0, output_padding=0)

    def forward(self, x):
        return self.block(x)

# blue box in Fig.1
class BlueBlock(nn.Module):
    def __init__(self, in_planes, out_planes):
        super().__init__()
        self.block = nn.Sequential(TranspConv3DBlock(in_planes, out_planes),
                                    Conv3DBlock(out_planes, out_planes,double=False))
    def forward(self, x):
        return self.block(x)


import torch.nn as nn
from einops import rearrange

from self_attention_cv.pos_embeddings import AbsPositionalEncoding1D


class Embeddings3D(nn.Module):
    def __init__(self, input_dim, embed_dim, cube_size, patch_size=16, dropout=0.1):
        super().__init__()
        self.n_patches = int((cube_size[0] * cube_size[1] * cube_size[2]) / (patch_size * patch_size * patch_size))
        self.patch_size = patch_size
        self.embed_dim = embed_dim
        self.patch_embeddings = nn.Conv3d(in_channels=input_dim, out_channels=embed_dim,
                                          kernel_size=patch_size, stride=patch_size, bias=False)
        self.position_embeddings = AbsPositionalEncoding1D(self.n_patches, embed_dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        """
        x is a 5D tensor
        """
        x = rearrange(self.patch_embeddings(x), 'b d x y z -> b (x y z) d')
        embeddings = self.dropout(self.position_embeddings(x))
        return embeddings



import codecs
import os
import pathlib
import platform
import re

from setuptools import setup, find_packages


def clean_html(raw_html):
    """
    Args:
        raw_html:
    Returns:
    """
    cleanr = re.compile("<.*?>")
    cleantext = re.sub(cleanr, "", raw_html).strip()
    return cleantext


# Single sourcing code from here:
# https://packaging.python.org/guides/single-sourcing-package-version/
def find_version(*file_paths):
    """
    Args:
        *file_paths:
    Returns:
    """
    here = os.path.abspath(os.path.dirname(__file__))

    def read(*parts):
        """
        Args:
            *parts:
        Returns:
        """
        with codecs.open(os.path.join(here, *parts), "r") as fp:
            return fp.read()

    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def fetch_long_description():
    """
    Returns:
    """
    with open("README.md", encoding="utf8") as f:
        readme = f.read()
        # https://stackoverflow.com/a/12982689
        readme = clean_html(readme)
    return readme


def fetch_requirements():
    """
    Returns:
    """
    requirements_file = "requirements.txt"

    if platform.system() == "Windows":
        DEPENDENCY_LINKS.append("https://download.pytorch.org/whl/torch_stable.html")

    with open(requirements_file) as f:
        reqs = f.read()

    reqs = reqs.strip().split("\n")
    return reqs


HERE = pathlib.Path(__file__).parent

DISTNAME = "self_attention_cv"
DESCRIPTION = "Self-attention building blocks for computer vision applications in PyTorch"

LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESCRIPTION_CONTENT_TYPE = "text/markdown"
URL = "https://github.com/The-AI-Summer/self_attention_cv"
AUTHOR = "Adaloglou Nikolas"
AUTHOR_EMAIL = "nikolas@theaiusummer.com"
LICENSE = "MIT"
DEPENDENCY_LINKS = []
REQUIREMENTS = (fetch_requirements())
EXCLUDES = ("examples")
EXT_MODULES = []

if __name__ == "__main__":
    setup(
        name=DISTNAME,
        install_requires=REQUIREMENTS,
        url=URL,
        license=LICENSE,
        include_package_data=True,
        version=find_version("self_attention_cv", "version.py"),
        packages=find_packages(exclude=EXCLUDES),
        python_requires=">=3.6",
        ext_modules=EXT_MODULES,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        dependency_links=DEPENDENCY_LINKS,
        classifiers=[
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
        ]
    )



from .abs_pos_emb1D import AbsPosEmb1D
from .abs_pos_encoding_1d import AbsPositionalEncoding1D
from .relative_embeddings_1D import rel_pos_emb_1d, RelPosEmb1D
from .relative_embeddings_2D import RelPosEmb2D
from .relative_pos_enc_qkv import Relative2DPosEncQKV
from .pos_encoding_sin import PositionalEncodingSin



import torch.nn as nn
from einops import rearrange

from .bottleneck_layer import Bottleneck
from .decoder import Up, SignleConv
from ..vit import ViT


class TransUnet(nn.Module):
    def __init__(self, *, img_dim, in_channels, classes,
                 vit_blocks=12,
                 vit_heads=4,
                 vit_dim_linear_mhsa_block=1024,
                 vit_transformer=None,
                 vit_channels = None
                 ):
        """
        My reimplementation of TransUnet based on the paper:
        https://arxiv.org/abs/2102.04306
        Badly written, many details missing and significantly differently
        from the authors official implementation (super messy code also :P ).
        My implementation doesnt match 100 the authors code.
        Basically I wanted to see the logic with vit and resnet backbone for
        shaping a unet model with long skip connections.

        Args:
            img_dim: the img dimension
            in_channels: channels of the input
            classes: desired segmentation classes
            vit_blocks: MHSA blocks of ViT
            vit_heads: number of MHSA heads
            vit_dim_linear_mhsa_block: MHSA MLP dimension
            vit_transformer: pass your own version of vit
            vit_channels: the channels of your pretrained vit. default is 128*8
        """
        super().__init__()
        self.inplanes = 128
        vit_channels = self.inplanes * 8 if vit_channels is None else vit_channels

        # Not clear how they used resnet arch. since the first input after conv
        # must be 128 channels and half spat dims.
        in_conv1 = nn.Conv2d(in_channels, self.inplanes, kernel_size=7, stride=2, padding=3,
                             bias=False)
        bn1 = nn.BatchNorm2d(self.inplanes)
        self.init_conv = nn.Sequential(in_conv1, bn1, nn.ReLU(inplace=True))
        self.conv1 = Bottleneck(self.inplanes, self.inplanes * 2, stride=2)
        self.conv2 = Bottleneck(self.inplanes * 2, self.inplanes * 4, stride=2)
        self.conv3 = Bottleneck(self.inplanes * 4, vit_channels, stride=2)

        self.img_dim_vit = img_dim // 16

        self.vit = ViT(img_dim=self.img_dim_vit,
                       in_channels=vit_channels,  # encoder channels
                       patch_dim=1,
                       dim=vit_channels,  # vit out channels for decoding
                       blocks=vit_blocks,
                       heads=vit_heads,
                       dim_linear_block=vit_dim_linear_mhsa_block,
                       classification=False) if vit_transformer is None else vit_transformer

        self.vit_conv = SignleConv(in_ch=vit_channels, out_ch=512)

        self.dec1 = Up(1024, 256)
        self.dec2 = Up(512, 128)
        self.dec3 = Up(256, 64)
        self.dec4 = Up(64, 16)
        self.conv1x1 = nn.Conv2d(16, classes, kernel_size=1)

    def forward(self, x):
        # ResNet 50-like encoder
        x2 = self.init_conv(x)  # 128,64,64
        x4 = self.conv1(x2)  # 256,32,32
        x8 = self.conv2(x4)  # 512,16,16
        x16 = self.conv3(x8)  # 1024,8,8
        y = self.vit(x16)
        y = rearrange(y, 'b (x y) dim -> b dim x y ', x=self.img_dim_vit, y=self.img_dim_vit)
        y = self.vit_conv(y)
        y = self.dec1(y, x8)  # 256,16,16
        y = self.dec2(y, x4)
        y = self.dec3(y, x2)
        y = self.dec4(y)
        return self.conv1x1(y)



import torch
from torch import nn, einsum


class AbsPosEmb1D(nn.Module):
    """
    Given query q of shape [batch heads tokens dim] we multiply
    q by all the flattened absolute differences between tokens.
    Learned embedding representations are shared across heads
    """

    def __init__(self, tokens, dim_head):
        """
        Output: [batch head tokens tokens]
        Args:
            tokens: elements of the sequence
            dim_head: the size of the last dimension of q
        """
        super().__init__()
        scale = dim_head ** -0.5
        self.abs_pos_emb = nn.Parameter(torch.randn(tokens, dim_head) * scale)

    def forward(self, q):
        return einsum('b h i d, j d -> b h i j', q, self.abs_pos_emb)



import torch.nn as nn
from torchvision.models import resnet50

from .vit import ViT


class ResNet50ViT(nn.Module):
    def __init__(self, *, img_dim, pretrained_resnet=False,
                 resnet_layers=5,
                 num_classes=10,
                 dim=512,
                 blocks=6,
                 heads=4,
                 dim_linear_block=1024,
                 dim_head=64,
                 dropout=0, transformer=None, classification=True
                 ):
        """
        ResNet50 + ViT for image classification
        Args:
            img_dim: the spatial image size
            pretrained_resnet: wheter to load pretrained weight from torch vision
            resnet_layers: use 5 or 6. the layer to keep from the resnet 50 backbone
            num_classes: classification task classes
            dim: the linear layer's dim to project the patches for MHSA
            blocks: number of transformer blocks
            heads: number of heads
            dim_linear_block: inner dim of the transformer linear block
            dim_head: dim head in case you want to define it. defaults to dim/heads
            dropout: for pos emb and transformer
            transformer: in case you want to provide another transformer implementation
        """
        super().__init__()
        assert 5 <= resnet_layers <= 6, f'Inser 5 or 6 resnet layers to keep'
        resnet_channels = 256 if resnet_layers == 5 else 512
        feat_dim = img_dim // 4 if resnet_layers == 5 else img_dim // 8

        resnet_layers = list(resnet50(pretrained=pretrained_resnet).children())[:resnet_layers]
        self.img_dim = img_dim

        res50 = nn.Sequential(*resnet_layers)

        vit = ViT(img_dim=feat_dim, in_channels=resnet_channels, patch_dim=1,
                  num_classes=num_classes, dim_linear_block=dim_linear_block, dim=dim,
                  dim_head=dim_head, dropout=dropout, transformer=transformer,
                  classification=classification, heads=heads, blocks=blocks)
        self.model = nn.Sequential(res50, vit)

    def forward(self, x):
        b, c, h, w = x.shape
        assert c == 3, f'Insert input with 3 inp channels, {c} received.'
        assert self.img_dim == h == w, f'Insert input with {self.img_dim} dimensions.'
        return self.model(x)
